#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include <linux/dvb/dmx.h>
#include <linux/dvb/frontend.h>

#include "getstream.h"
#include "dvbrc.h"
#include "log.h"

/*
 * Dump FrontEnd Status byte as clear text returned
 * by GET_STATUS or GET_EVENT ioctl
 *
 */
char * fe_decode_status(int status) {
	static char	str[256];

	str[0]=0x0;

	if (status & FE_HAS_SIGNAL) 
		strcat(str, "HAS_SIGNAL ");
	if (status & FE_HAS_CARRIER)
		strcat(str, "HAS_CARRIER ");
	if (status & FE_HAS_VITERBI) 
		strcat(str, "HAS_VITERBI ");
	if (status & FE_HAS_SYNC) 
		strcat(str, "HAS_SYNC ");
	if (status & FE_HAS_LOCK) 
		strcat(str, "HAS_LOCK ");
	if (status & FE_TIMEDOUT) 
		strcat(str, "TIMEDOUT ");
	if (status & FE_REINIT) 
		strcat(str, "REINIT ");

	return str;
}

/*
 * Wait for the frontend to deliver an event
 * Return to callee if either TIMEDOUT or HAS_LOCK
 *
 */
int fe_get_event(int fd) {
	struct pollfd pfd;
	struct dvb_frontend_event event;
	int	res;

	pfd.fd = fd;
	pfd.events = POLLIN;

	do {
		if (!poll(&pfd, 1, 10000))
			continue;

		if (!(pfd.revents & POLLIN))
			continue;

		res = ioctl(fd, FE_GET_EVENT, &event);

		/* EOVERFLOW happens if we didnt poll for events
		 * lately and the event queue (8) overflowed inbetween. */
		if (res < 0 && errno != EOVERFLOW) {
			perror("fatal ERROR on ioctl() FE_GET_EVENT");
			exit(-1);
		}

		if (res >= 0 && event.status) {
			putlog(0, "Status: %02x %s", event.status, fe_decode_status(event.status));
			if ((console_loglevel > 0) || (file_loglevel > 0)) {
				uint16_t sigstr;
				uint16_t snr; /* Signal Noise Ratio */
				uint32_t ber; /* Bit Error Rate */
				if (ioctl(fd, FE_READ_SIGNAL_STRENGTH, &sigstr) < 0) sigstr = 0;
				if (ioctl(fd, FE_READ_SNR, &snr) < 0) snr = 0;
				if (ioctl(fd, FE_READ_BER, &ber) < 0) ber = 0;
				putlog(1, "Signal strength: %d   SNR: %3.0f%%   BER: %lx",
				       (unsigned int)sigstr, (double)snr / 653.35, (unsigned long)ber);
			}
		}

	} while (((event.status & FE_HAS_LOCK) == 0) && ((event.status & FE_TIMEDOUT) == 0));

	return(event.status);
}

int fe_tune(int adapter, struct transponder * t) {

	struct dvb_frontend_parameters	feparams;
	struct lnb_s			*lnb;
	char				fename[128];
	int 				fdf, status;
	fe_sec_tone_mode_t		tone;
	fe_sec_voltage_t		voltage;

	memset(&feparams, 0, sizeof(struct dvb_frontend_parameters));

	sprintf(fename, "/dev/dvb/adapter%d/frontend0", adapter);

	fdf = open(fename, O_RDWR);

	if (!fdf) {
		perror("Error opening dvb frontend");
		exit(-1);
	}

	if (t->type == DVB_S) {
	
		/* Select vertical (13 V) or horizontal polarity (18 V) */
		voltage = (t->tp.s.pol == 'H') ? SEC_VOLTAGE_18 : SEC_VOLTAGE_13;
	
		if (ioctl(fdf, FE_SET_VOLTAGE, voltage) < 0) {
			perror("fatal ERROR on ioctl() FE_SET_VOLTAGE");
			putlog(5, "This error can be caused by trying to use a DVB-T or DVB-C card for DVB-S");
			exit(-1);
		}
	
		lnb = t->tp.s.satellite->lnb;
	
		if (t->tp.s.freq > 2200000) { /* Frequency contains l.osc.f. */
			if (t->tp.s.freq < lnb->slof) {
				feparams.frequency = (t->tp.s.freq - lnb->lof1);
				tone = SEC_TONE_OFF;
			} else {
				feparams.frequency = (t->tp.s.freq - lnb->lof2);
				tone = SEC_TONE_ON;
			}
		} else {
			feparams.frequency = t->tp.s.freq;
		}
	
		feparams.inversion = INVERSION_AUTO;
		feparams.u.qpsk.symbol_rate = t->tp.s.srate;
		feparams.u.qpsk.fec_inner = FEC_AUTO;
	
		if (ioctl(fdf, FE_SET_TONE, tone) < 0) {
			perror("fatal ERROR on ioctl() FE_SET_TONE");
			exit(-1);
		}

		putlog(3, "FE_SET_FRONTEND:");
		putlog(3, "\ttone = %d", tone);
		putlog(3, "\tvoltage = %d", voltage);
		putlog(3, "\tt->tp.s.freq  = %lu", t->tp.s.freq);
		putlog(3, "\tt->tp.s.srate = %lu", t->tp.s.srate);
		putlog(3, "\tlnb->lof1 = %lu", lnb->lof1);
		putlog(3, "\tlnb->lof2 = %lu", lnb->lof2);
		putlog(3, "\tlnb->slof = %lu", lnb->slof);
		putlog(3, "\tfeparams.frequency = %d", feparams.frequency);
		putlog(3, "\tfeparams.inversion = %d", feparams.inversion);
		putlog(3, "\tfeparams.u.qpsk.symbol_rate = %d", feparams.u.qpsk.symbol_rate);
		putlog(3, "\tfeparams.u.qpsk.fec_inner = %d", feparams.u.qpsk.fec_inner);

	} else if (t->type == DVB_T) {

		/* foo */
		feparams.inversion = INVERSION_AUTO;
		feparams.frequency = t->tp.t.freq;
	
		if        (t->tp.t.bandw == 0) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_AUTO;
		} else if (t->tp.t.bandw == 6) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
		} else if (t->tp.t.bandw == 7) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
		} else if (t->tp.t.bandw == 8) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
		} else {
			putlog(0, "WARNING: Unknown bandwidth %u", t->tp.t.bandw);
		}
	
		/* FIXME make this configurable */
		feparams.u.ofdm.code_rate_HP = FEC_AUTO;
		feparams.u.ofdm.code_rate_LP = FEC_AUTO;
	
		if        (t->tp.t.modul ==   0) {
			feparams.u.ofdm.constellation = QAM_AUTO;
		} else if (t->tp.t.modul ==  16) {
			feparams.u.ofdm.constellation = QAM_16;
		} else if (t->tp.t.modul ==  32) {
			feparams.u.ofdm.constellation = QAM_32;
		} else if (t->tp.t.modul ==  64) {
			feparams.u.ofdm.constellation = QAM_64;
		} else if (t->tp.t.modul == 128) {
			feparams.u.ofdm.constellation = QAM_128;
		} else if (t->tp.t.modul == 256) {
			feparams.u.ofdm.constellation = QAM_256;
		} else {
			putlog(0, "WARNING: Unknown modulation %u", t->tp.t.modul);
		}
	
		if        (t->tp.t.trans == 0) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;
		} else if (t->tp.t.trans == 2) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_2K;
		} else if (t->tp.t.trans == 8) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_8K;
		} else {
			putlog(0, "WARNING: Unknown transmission mode %u", t->tp.t.trans);
		}
	
		if        (t->tp.t.guard ==  0) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
		} else if (t->tp.t.guard ==  4) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_4;
		} else if (t->tp.t.guard ==  8) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_8;
		} else if (t->tp.t.guard == 16) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_16;
		} else if (t->tp.t.guard == 32) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_32;
		} else {
			putlog(0, "WARNING: Unknown guard interval %u", t->tp.t.guard);
		}
	
		if        (t->tp.t.hier == -1) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_NONE;
		} else if (t->tp.t.hier ==  0) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_AUTO;
		} else if (t->tp.t.hier ==  1) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_1;
		} else if (t->tp.t.hier ==  2) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_2;
		} else if (t->tp.t.hier ==  4) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_4;
		} else {
			putlog(0, "WARNING: Unknown hierarchy %u", t->tp.t.hier);
		}

/*	} else if (t->type == DVB_C) { */ /* Implement me! */
	} else {
	
		fprintf(stderr, "Unknown/unimplemented DVB Transponder type: I have no idea how to tune this.\n");
		exit(-1);

	}

	if (ioctl(fdf, FE_SET_FRONTEND, &feparams) < 0) {
		perror("fatal ERROR on ioctl() FE_SET_FRONTEND");
		exit(-1);
	}

	do {
		status = fe_get_event(fdf);
	} while (!(status && FE_HAS_LOCK) && !(status && FE_TIMEDOUT));

	/* TODO We should return the condition whether we suceeded */

	return 1;
}

